#pragma once
namespace uncrustify
{

namespace limits
{

static constexpr int MAX_OPTION_NAME_LEN = 32;
static constexpr int AL_SIZE             = 8000;

} // namespace limits

} // namespace uncrustify
