methodCall([] {
                             Log();
                           });

funcCall(match ( [ & ]( const ContentProps &props)  {
        return PairingCmpnt()
      }));

match( [ & ]( const ContentProps &props)  {
        return PairingCmpnt()
      });


outerMethodCall(methodCall(^{
  // action
}, x)
);

outerMethodCall(x, methodCall(^{
  // action
}, y));

options({
          .cornerRadius = CGFLOAT_MAX,
        });

mapToPtr([&](const LeftAddOn::Props &addOnProps) {
      FSTheme *const theme = AK::getTheme();
});

mapToPtr(x, [&](const Props &addOnProps) {
      FSTheme *const theme = AK::getTheme();
});

mapToPtr([&](const Props &addOnProps) {
      FSTheme *const theme = AK::getTheme();
});

methodCall(arg1, arg2,
arg3);

methodCall(arg1, arg2, arg3);

methodCall(arg1, []{
              variant.action.send(Cmpnt);
      }, arg3);

methodCall(arg1, {
              .x = 10,
      }, arg3);

methodCall({
              .x = 10,
      }, arg3);

methodCall(arg1, {
              .x = 10,
      });
